/*
 * Copyright (c) 2016, 2020, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and 
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

package com.oracle.iot.client.impl.http;

import android.webkit.CookieManager;
import com.oracle.iot.client.RestApi;
import com.oracle.iot.client.impl.util.Base64;
import oracle.iot.client.enterprise.UserAuthenticationException;

import java.net.HttpURLConnection;
import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocketFactory;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.GeneralSecurityException;

/**
 * Platform specific details for HttpClient
 */
class HttpClientBase {

	private static final String CSRF_TOKEN_REQUEST_PATH = "/iot/webapi/v2/private/server";
	private static final String CSRF_TOKEN_NAME = "X-CSRF-TOKEN";

		protected void beforeConnect(HttpURLConnection connection)
				throws IOException, GeneralSecurityException {

			if (!RestApi.V2.isWebApi()) return;

			final String csrfToken = getCsrfToken(connection);

			if(csrfToken != null)
				connection.setRequestProperty(CSRF_TOKEN_NAME, csrfToken);

			// Add Cookies
			final String cookie = CookieManager.getInstance().getCookie(connection.getURL().toString());

			if (cookie.contains(CSRF_TOKEN_NAME))
				connection.setRequestProperty("Cookie", cookie);
			else if (csrfToken != null)
				connection.setRequestProperty("Cookie", cookie + "; " + CSRF_TOKEN_NAME + "=" + csrfToken);
		}

		private String getCsrfToken(HttpURLConnection connection)
				throws IOException, GeneralSecurityException {
			URL url = connection.getURL();
			URL csrfTokenUrl = new URL("https", url.getHost(), url.getPort(), CSRF_TOKEN_REQUEST_PATH);
			final HttpsURLConnection con = (HttpsURLConnection) csrfTokenUrl.openConnection();

			con.setSSLSocketFactory(((HttpsURLConnection)connection).getSSLSocketFactory());
			con.setInstanceFollowRedirects(false);
			con.setConnectTimeout(getResponseTimeout());
			con.setReadTimeout(getResponseTimeout());
			con.setRequestMethod("GET");

			String cookie = CookieManager.getInstance().getCookie(con.getURL().toString());
			con.setRequestProperty("Cookie", cookie);

			con.connect();

			final int responseCode = con.getResponseCode();
			String csrfToken = null;
			if (responseCode == 200) {
        csrfToken = con.getHeaderField(CSRF_TOKEN_NAME);

        // Validate that the CSRF token is valid BASE 64 data.
        Base64.Decoder base64Decoder = Base64.getDecoder();

        try {
            byte[] csrfTokenBytes = base64Decoder.decode(csrfToken);
        } catch(IllegalArgumentException iae) {
            throw new GeneralSecurityException("CSRF token is invalid.");
        }
			} else if (responseCode == 302) {
				throw new UserAuthenticationException("User Authentication failed!", con.getHeaderField("Location"));
			}

			con.disconnect();
			return csrfToken;
		}

    private static final int DEFAULT_RESPONSE_TIMEOUT = 15000; // milliseconds
    private static final int responseTimeout;
    static {
        int value = DEFAULT_RESPONSE_TIMEOUT;
        try {
            value = Integer.getInteger(
                "oracle.iot.client.http_response_timeout",
                DEFAULT_RESPONSE_TIMEOUT);
        } catch (SecurityException e) {
            // use default value
            value = DEFAULT_RESPONSE_TIMEOUT;
        } finally {
            responseTimeout = value > 0 ? value : 0;
        }
    }

    private final URL url;

    HttpClientBase(URL url) {
        this.url = url;
    }

    /**
     * Get the HTTP response timeout.
     * The timeout The default value is 15000 milliseconds and can be set
     * with the {@code oracle.iot.client.http_response_timeout} property.
     * @return the HTTP response timeout
     */
    protected int getResponseTimeout() {
        return responseTimeout;
    }

    protected URL getUrl() {
        return url;
    }

}
